/**
 * VampirTrace Plugin Counter for PowerTracer
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, ZIH, TU Dresden, Federal Republic of Germany
 *
 * See the file COPYING in the package base directory for details
 **/

/*
libeeClustPlugin.so,
a library to include post-mortem events from DB on eeClust master into traces gnerated by VampirTrace
Copyright (C) 2010 TU Dresden, ZIH
*/

//TODO? compensate delay of DB timestamps

/* some standard functions */
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <glib.h>

#include <vampirtrace/vt_plugin_cntr.h> /* VT plugin interface definition */
#include <DBConnector.h>                /* DB connection API */

#define PL_SUCCESS      0
#define E_NO_INIT      -1
#define E_INIT_FAILURE -2
#define E_INVALID_ARG  -3
#define E_NOT_FOUND    -4

/* counters are solely identified by their name (== name of table in DB) */
typedef struct db_counter {
  char* name;
  char* metric;
  char* node;
  char* unit;
} db_counter_t;

static db_counter_t *db_counters = NULL;
static int init_failed = E_NO_INIT;  //TODO return error if any function is called and init_failed!=0
static int num_counters = 0;

/* time in usec*/
GTimeVal timestamp;

/* nedded to calculate VampirTrace timestamp from internal measurement */
static uint64_t vt_timer_start;
static uint64_t vt_timer_stop;
static uint64_t internal_timer_start;
static uint64_t internal_timer_stop;
double factor=0.0,additional=0.0;
uint64_t diff;

/**
 * internal helper functions
 */

/*
 * check if a string matches a pattern that can include "*" and "?" as wildcards
 * @param ptrn[in] pattern
 * @param name[in] string that is checked if it matches the given pattern
 * @return 0 if name matches pattern, -1 elsewise
 */
static int strwldcmp(char * ptrn, char * name)
{
  char ** parts=NULL, ** substrings=NULL;
  char * pattern, * cand;
  char * pos=NULL, *str_pos=NULL;
  int num_parts=0,num_substrings;
  int i,j;
  int ret=0;

  if ((ptrn==NULL)||(strlen(ptrn)==0)||(name==NULL)) return E_INVALID_ARG;
  
  pattern=strdup(ptrn);
  cand=name;
  
  /* divide pattern into substrings divided by "*" */
  pos=pattern;
  do
  {
    num_parts++;
    parts=(char**) realloc(parts,num_parts*sizeof(char*));
    parts[num_parts-1]=pos;
    pos=strstr(pos,"*");
    if (pos!=NULL) {*pos='\0';pos++;}
  }
  while (pos!=NULL);

  /* try to find all parts in name */
  for(i=0;i<num_parts;i++)
  {
    pos=parts[i];
    /* handle empty parts */
    if ((i==0)&&(strlen(parts[i])==0)) continue;        // first part empty => always ok
    if ((i==num_parts-1)&&(strlen(parts[i])==0)) break; // last part empty => rest of characters consumed by "*"
    
    /* divide part into substrings divided by "?" */
    num_substrings=0;
    do
    {
      num_substrings++;
      substrings=(char**)realloc(substrings,num_substrings*sizeof(char*));
      substrings[num_substrings-1]=pos;
      pos=strstr(pos,"?");
      if (pos!=NULL) {*pos='\0';pos++;}
    }
    while (pos!=NULL);
    
    pos=cand-1;
    /* search substrings in name*/
    do
    {
     str_pos=pos;pos++;
     if ((i==0)&&(pos!=cand)) {ret=-1;break;}
     /* find substring */
     for(j=0;j<num_substrings;j++)
     {
      str_pos++;
      if (strlen(substrings[j])==0) continue;
      if (strlen(str_pos)<strlen(substrings[j])) {ret=-1;break;}
      if (strncmp(str_pos,substrings[j],strlen(substrings[j]))) break;
      else str_pos+=strlen(substrings[j]); 
     }
     /* substring found */
     if (j==num_substrings){
       if ((i==num_parts-1)&&(strlen(pos)!=str_pos-pos)) continue; // last part has to consume all remaining characters
       else
       {
         cand+=strlen(parts[i]);
         break;
       }
     }
    }
    while (strlen(pos)>0);
    if (ret==-1) break;
  }
  
  free(pattern);
  free(parts);
  free(substrings);
  
  return ret;
}

/**
 * plugin interface implementation
 */

/* VT timer function */
static uint64_t (*vt_timer_function)(void) = NULL;

/* set the reference timer used by VampirTrace */
void set_timer_function(uint64_t(*timestamp)(void))
{
  vt_timer_function=timestamp;
}

/* initialize plugin 
 * @return PL_SUCCSESS if plugin is initialized successfully, E_INIT_FAILURE elsewise
 */
int32_t pl_start(void)
{
  char ** events=NULL;
  uint64_t tmp1,tmp2;
  int i,count;
  
  if (vt_timer_function==NULL){
    init_failed=E_INIT_FAILURE;
    fprintf(stderr,"received invalid timer function from VampirTrace\n");
  }
  else{
    init_failed=PL_SUCCESS;
    /* start timestamp synchronisation */
    diff=0xffffffffffffffff;
    for(i=0;i<100;i++)
    {
     tmp1=vt_timer_function();
     g_get_current_time(&timestamp);
     tmp2=vt_timer_function();
     if (tmp2-tmp1<diff){ 
       diff=tmp2-tmp1;
       internal_timer_start=((uint64_t)timestamp.tv_sec)*1000000+(uint64_t)timestamp.tv_usec;
       vt_timer_start=tmp1+diff/2;
     }
    }
    
    /* read available events from db */
    count=db_getMetrics(&events);
    if (count>0) num_counters=count;

#ifdef DEBUG
    printf("%d different metrics available in db\n", num_counters);
#endif

    db_counters=calloc(num_counters,sizeof(db_counter_t));
    while (count>0){
      count--;
      db_counters[count].name=strdup(events[count]);
      char * pch = strtok (events[count],"_");
      db_counters[count].node = strdup(pch);
      pch = strtok (NULL,"_");
      db_counters[count].metric = strdup(db_counters[count].name + strlen(db_counters[count].node) + 1);
      if(strcmp(db_counters[count].metric,"power") == 0)
	db_counters[count].unit = "watt";
      else if(strstr(db_counters[count].metric,"bw") != 0)
	db_counters[count].unit = "MB/s";
      else if(strstr(db_counters[count].metric,"state") != 0 && strstr(db_counters[count].metric,"core") != 0)
	db_counters[count].unit = "kHz";
      else if(strstr(db_counters[count].metric,"state") != 0 && strstr(db_counters[count].metric,"nic") != 0)
	db_counters[count].unit = "MBit";
      else
	db_counters[count].unit = "no";

#ifdef DEBUG
      printf("Created new counter %s with node %s, metric %s and unit %s.\n", db_counters[count].name, db_counters[count].node, db_counters[count].metric, db_counters[count].unit);
#endif
    }
  }
  return init_failed;
}

/* get info about counters that match the specified pattern
 * @param event_name[in] pattern ("*" and "?" wildcards supported)
 * @return array of events that match the given pattern
 */
vt_plugin_cntr_metric_info * get_event_info(char * event_name)
{
  int count,selected=0;
  vt_plugin_cntr_metric_info * return_values;

  #ifdef DEBUG
  printf("get information about events for pattern \"%s\"\n",event_name);fflush(stdout);
  #endif
   
  return_values = malloc(sizeof(vt_plugin_cntr_metric_info));

  count=num_counters;
  while (count>0)
  {
   count--; 
   //#ifdef DEBUG
   //printf("  %3i: %12s ", count, db_counters[count].name);
   //#endif

   if (!strwldcmp(event_name,db_counters[count].name))
   { 
     //#ifdef DEBUG
     //printf("- match\n"); 
     //#endif
     selected++;
     return_values = realloc(return_values,(selected+1) * sizeof(vt_plugin_cntr_metric_info));
     return_values[selected-1].name = strdup(db_counters[count].name);
     //TODO get unit from db or config file
     return_values[selected-1].unit = strdup(db_counters[count].unit);
     return_values[selected-1].cntr_property = VT_PLUGIN_CNTR_ABS|VT_PLUGIN_CNTR_DOUBLE|VT_PLUGIN_CNTR_LAST;
   }
   //#ifdef DEBUG
   //else printf("- failed\n");
   //#endif
  }
  
  /* Last element empty */
  return_values[selected].name = NULL;

  return return_values;
}                                                    

/* activate a certain counter 
 * @param event_name[in] name of counter that should be recorded
 * @return a unique ID used by VampiTrace to collect the data later on, E_NOT_FOUND in case of an error
 */
int32_t add_counter(char * event_name)
{
  int i;
  #ifdef DEBUG
  printf("Adding counter \"%s\"\n",event_name);fflush(stdout);
  #endif
  for (i=0;i<num_counters;i++)  { 
    /* data collection in db is allways running, thus no starting of counters required
     * only assign event name to returned ID
     */
    #ifdef DEBUG
    if (!strcmp(db_counters[i].name,event_name)) printf("  ID: %i\n",i);
    #endif
    if (!strcmp(db_counters[i].name,event_name)) return i;
  }
  fprintf(stderr,"event %s not found\n", event_name);
  return E_NOT_FOUND; 
}

/* return all values for a counter
 * @param counter[in] ID of counter that should be read from db
 * @param result_vector[out] array of (timestamp,value) pairs in vt_plugin format
 * @return number of returned array elements, E_NO_INIT or E_INIT_FAILURE if called uninitialized
 */
uint64_t get_all_values(int32_t counter, vt_plugin_cntr_timevalue ** result_vector)
{
 row_struct * raw_data;
 vt_plugin_cntr_timevalue * results;
 uint64_t tmp1,tmp2;
 int i,count;

 union{
   uint64_t u64;
   double dbl;
 }value;
 
 /* no start timestamp and timer function available if initialization failed */
 if (init_failed!=PL_SUCCESS) return init_failed;

 #ifdef DEBUG
 printf("get values for counter %i: \"%s\" ...\n",counter,db_counters[counter].name);fflush(stdout);
 #endif

 /* stop timestamp synchronisation */
 diff=0xffffffffffffffff;
 if (((int)factor)==0) // calculate adjustment parameters only once
 {
  for(i=0;i<100;i++) 
   {
    tmp1=vt_timer_function();
    g_get_current_time(&timestamp);
    tmp2=vt_timer_function();
    if (tmp2-tmp1<diff){
      diff=tmp2-tmp1;
      internal_timer_stop=((uint64_t)timestamp.tv_sec)*1000000+(uint64_t)timestamp.tv_usec;
      vt_timer_stop=tmp1+diff/2;
    }
   }
  /* determine timestamp adjustment parameters */
  factor= ((double)(vt_timer_stop-vt_timer_start))/((double)(internal_timer_stop-internal_timer_start));
  additional=((double)vt_timer_start)-factor*((double)internal_timer_start);       
 }
   
 /* read values from db */
 count=db_getNodeMetric(db_counters[counter].metric, db_counters[counter].node, internal_timer_start, internal_timer_stop, &raw_data);
 
 #ifdef DEBUG
 printf(" received %i values from db for %5llu ms\n",count,(internal_timer_stop-internal_timer_start)/1000);fflush(stdout);
 #endif

 if(count == 0)
 {
  results=(vt_plugin_cntr_timevalue*)malloc(sizeof(vt_plugin_cntr_timevalue));
  results[0].timestamp = 0;
  results[0].value = 0;

  /* update pointer to results */
  *result_vector=results;    
  return 0;
 } 
 if (count>0)
 {
  /* convert results in format required by VampirTrace */
  results=(vt_plugin_cntr_timevalue*)malloc((count+1)*sizeof(vt_plugin_cntr_timevalue));
//  results=(vt_plugin_cntr_timevalue*)malloc(count*sizeof(vt_plugin_cntr_timevalue));
  results[count].timestamp = 0;
  results[count].value = 0;
  #ifdef DEBUG
  printf("start: %llu\n",internal_timer_start);
  printf("stop: %llu\n",internal_timer_stop);
  #endif
  for(i=0;i<count;i++)
  {
    results[i].timestamp=additional+factor*raw_data[i].timestamp;
    value.dbl=raw_data[i].value;
    results[i].value=value.u64;
  }
  free(raw_data);
  
  /* update pointer to results */
  *result_vector=results;    

  return count;
 }

 return 0;
}

/* free resources */
void pl_stop(void)
{  
  db_finishConnection();
  free(db_counters);
}

/* Plugin description needed by VampirTrace
 * @return plugin property struct 
 */
vt_plugin_cntr_info get_info()
{
  vt_plugin_cntr_info info;
  memset(&info,0,sizeof(vt_plugin_cntr_info));
  
  info.set_pform_wtime_function = set_timer_function;
  info.init                     = pl_start;
  info.add_counter              = add_counter;
                   
  info.vt_plugin_cntr_version   = VT_PLUGIN_CNTR_VERSION;
  info.run_per                  = VT_PLUGIN_CNTR_ONCE;                // not process or thread specific
  info.synch                    = VT_PLUGIN_CNTR_ASYNCH_POST_MORTEM;  // merge values post mortem
  info.get_event_info           = get_event_info;
  info.get_all_values           = get_all_values;
  info.finalize                 = pl_stop;
  return info;
}
                                                
