/**
 * VampirTrace Plugin Counter for PowerTracer
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, ZIH, TU Dresden, Federal Republic of Germany
 *
 * See the file COPYING in the package base directory for details
 **/

/* C Example */
#include <stdio.h>
#include <mpi.h>
#include <sys/time.h>


void all_sleep(int interval, int rank)
{
  if (rank == 0) {
    printf("everyone sleeping %d s\n", interval);
  }
  sleep(interval);

  MPI_Barrier(MPI_COMM_WORLD);
}

void master_sleep(int interval, int rank)
{
  if (rank == 0) {
    printf("only master sleeping %d s, others barrier\n", interval);
    sleep(interval);
  }

  MPI_Barrier(MPI_COMM_WORLD);
}

void busy_wait(int interval, int rank)
{
   double start;
   struct timeval time;
   gettimeofday( &time, (void *) 0);
   if (rank == 0) {
     printf("busy waitingn %d s\n", interval);
   }
   start = (double)time.tv_sec+(((double)time.tv_usec)/1000000);
   while ((double)time.tv_sec+(((double)time.tv_usec)/1000000) - start < (double)interval) {
     gettimeofday( &time, (void *) 0);
   }
   MPI_Barrier(MPI_COMM_WORLD);
}

int main (int argc, char* argv[])
{
  int rank, size;
  char buf[60];

  gethostname(buf, 60);
  MPI_Init(&argc, &argv);	/* starts MPI */
  MPI_Comm_rank (MPI_COMM_WORLD, &rank);	/* get current process id */
  MPI_Comm_size (MPI_COMM_WORLD, &size);	/* get number of processes */

  if (rank == 0) {
    printf("mpipattern with %d procs started.\n", size);
  }

  all_sleep(1, rank);
  master_sleep(1, rank);
  all_sleep(1, rank);
  master_sleep(1, rank);
  all_sleep(1, rank);
  busy_wait(1, rank);
  all_sleep(1, rank);
  busy_wait(1, rank);
  all_sleep(1, rank);

  MPI_Finalize();
  return 0;
}
